from __future__ import annotations

import xarray as xr

from magnify.registry import components


class Stitcher:
    def __init__(self, overlap: int = 102):
        self.overlap = overlap

    def __call__(self, assay: xr.Dataset, overlap=102) -> xr.Dataset:
        self.overlap = overlap
        if "tile_row" not in assay.dims and "tile_col" not in assay.dims:
            return assay
        elif "tile_row" not in assay.dims:
            assay = assay.expand_dims("tile_row", 2)
        elif "tile_col" not in assay.dims:
            assay = assay.expand_dims("tile_col", 3)

        tiles = assay.image[..., : -self.overlap, : -self.overlap]
        # Move the time and channel axes last so we can focus on joining images.
        tiles = tiles.transpose("tile_row", "tile_col", "im_row", "im_col", "channel", "time")
        tiles = xr.concat(tiles, dim="im_row")
        images = xr.concat(tiles, dim="im_col")
        # Move the time and channel axes back to the front.
        images = images.transpose("channel", "time", "im_row", "im_col")
        assay = assay.drop_vars("image")
        assay["image"] = images
        return assay

    @components.register("stitch")
    def make():
        return Stitcher()



from __future__ import annotations

import logging

from numpy.typing import ArrayLike
import cv2 as cv
import dask.array as da
import numpy as np
import scipy
import tqdm
import xarray as xr

from magnify import utils
import magnify.registry as registry

logger = logging.getLogger(__name__)


class ButtonFinder:
    def __init__(
        self,
        row_dist: float = 126.3,
        col_dist: float = 233.2,
        min_button_radius: int = 4,
        max_button_radius: int = 15,
        cluster_penalty: float = 10,
        region_length: int = 61,
    ):
        self.row_dist = row_dist
        self.col_dist = col_dist
        self.min_button_radius = min_button_radius
        self.max_button_radius = max_button_radius
        self.cluster_penalty = cluster_penalty
        self.region_length = region_length

    def __call__(
        self,
        assay: xr.Dataset,
        row_dist: float = 126.3,
        col_dist: float = 233.2,
        min_button_radius: int = 4,
        max_button_radius: int = 15,
        cluster_penalty: float = 10,
        region_length: int = 61,
        progress_bar: bool = False,
    ) -> xr.Dataset:
        self.row_dist = row_dist
        self.col_dist = col_dist
        self.min_button_radius = min_button_radius
        self.max_button_radius = max_button_radius
        self.cluster_penalty = cluster_penalty
        self.region_length = region_length

        num_rows, num_cols = assay.id.shape
        if isinstance(assay.search_channel, str):
            if assay.search_channel in assay.channel:
                search_channels = [assay.search_channel]
            elif assay.search_channel == "all":
                search_channels = assay.channel
            else:
                raise ValueError(f"{assay.search_channel} is not a channel name.")
        else:
            # We're searching across multiple channels.
            search_channels = assay.search_channel
        min_button_dist = round(min(self.row_dist, self.col_dist) / 2)
        if min_button_dist % 2 == 0:
            # Certain opencv functions require an odd blocksize.
            min_button_dist -= 1

        # Create the array of subimage regions.
        regions = da.empty(
            (
                num_rows,
                num_cols,
                assay.dims["channel"],
                assay.dims["time"],
                self.region_length,
                self.region_length,
            ),
            dtype=assay.image.dtype,
            chunks=(
                1,
                1,
                assay.dims["channel"],
                assay.dims["time"],
                self.region_length,
                self.region_length,
            ),
        )
        assay = assay.assign(
            region=(
                ("marker_row", "marker_col", "channel", "time", "row", "col"),
                regions,
            ),
            fg=(
                ("marker_row", "marker_col", "channel", "time", "row", "col"),
                da.empty_like(
                    regions,
                    dtype=bool,
                ),
            ),
            bg=(
                ("marker_row", "marker_col", "channel", "time", "row", "col"),
                da.empty_like(
                    regions,
                    dtype=bool,
                ),
            ),
        )
        # Create the x and y coordinates arrays for each button.
        assay = assay.assign_coords(
            x=(
                ["marker_row", "marker_col", "time"],
                np.empty((num_rows, num_cols, assay.dims["time"])),
            ),
            y=(
                ["marker_row", "marker_col", "time"],
                np.empty((num_rows, num_cols, assay.dims["time"])),
            ),
        )

        # Run the button finding algorithm for each timestep.
        for t, time in enumerate(tqdm.tqdm(assay.time, disable=not progress_bar)):
            # Preload all images for this timnepoint so we only read from disk once.
            images = assay.image.sel(time=time).to_numpy()
            points = np.empty((0, 2))
            for channel in search_channels:
                c = np.where(assay.channel == channel)[0][0]
                image = utils.to_uint8(images[c])
                # Step 1: Find an imperfect button mask by thresholding.
                mask = cv.adaptiveThreshold(
                    image,
                    maxValue=255,
                    adaptiveMethod=cv.ADAPTIVE_THRESH_GAUSSIAN_C,
                    thresholdType=cv.THRESH_BINARY,
                    blockSize=min_button_dist,
                    C=-1,
                )
                # Step 2: Find connected components and filter out points.
                _, _, stats, new_points = cv.connectedComponentsWithStats(mask, connectivity=4)

                # Ignore the background point.
                new_points = new_points[1:]
                stats = stats[1:]

                # Exclude large and small blobs.
                new_points = new_points[
                    (stats[:, cv.CC_STAT_HEIGHT] <= 2 * self.max_button_radius)
                    & (stats[:, cv.CC_STAT_WIDTH] <= 2 * self.max_button_radius)
                    & (stats[:, cv.CC_STAT_HEIGHT] >= 2 * self.min_button_radius)
                    & (stats[:, cv.CC_STAT_WIDTH] >= 2 * self.min_button_radius)
                ]
                # Remove points too close to other points in this channel.
                dist_matrix = np.linalg.norm(
                    new_points[np.newaxis] - new_points[:, np.newaxis], axis=2
                )
                dist_matrix[np.diag_indices(len(dist_matrix))] = np.inf
                new_points = new_points[np.min(dist_matrix, axis=1) > min_button_dist]

                if len(points) > 0:
                    # Remove points too close to other points in previous channels.
                    dist_matrix = np.linalg.norm(
                        points[np.newaxis] - new_points[:, np.newaxis], axis=2
                    )
                    new_points = new_points[np.min(dist_matrix, axis=1) > min_button_dist]

                # Add the new points to the list of seen points.
                points = np.concatenate([points, new_points])

            # Split the points into x and y components.
            x = points[:, 0]
            y = points[:, 1]

            # Step 3: Cluster the points into distinct rows and columns.
            # The number of buttons we expect to see per row/col can vary if we have blank buttons.
            points_per_row = (assay.id != "").sum(dim="marker_col")
            row_labels = cluster_1d(
                y,
                total_length=image.shape[0],
                num_clusters=num_rows,
                cluster_length=self.row_dist,
                ideal_num_points=points_per_row,
                penalty=self.cluster_penalty,
            )
            points_per_col = (assay.id != "").sum(dim="marker_row")
            col_labels = cluster_1d(
                x,
                total_length=image.shape[1],
                num_clusters=num_cols,
                cluster_length=self.col_dist,
                ideal_num_points=points_per_col,
                penalty=self.cluster_penalty,
            )

            # Exclude boundary points that didn't fall into clusters.
            in_cluster = (row_labels >= 0) & (col_labels >= 0)
            x, y = x[in_cluster], y[in_cluster]
            col_labels, row_labels = col_labels[in_cluster], row_labels[in_cluster]

            # Step 4: Draw lines through each cluster.
            row_slope, row_intercepts = regress_clusters(
                x, y, labels=row_labels, num_clusters=num_rows, ideal_num_points=points_per_row
            )
            # We treat column indices as y and row indices as x to avoid near-infinite slopes.
            col_slope, col_intercepts = regress_clusters(
                y,
                x,
                labels=col_labels,
                num_clusters=num_cols,
                ideal_num_points=points_per_col,
            )

            # Step 5: Set button locations as the intersection of each line pair.
            assay.y[..., t] = (
                row_slope * col_intercepts[np.newaxis] + row_intercepts[:, np.newaxis]
            ) / (1 - row_slope * col_slope)
            assay.x[..., t] = assay.y[..., t] * col_slope + col_intercepts[np.newaxis]

            # Step 6: Extract a region around each button.
            offsets = np.empty((num_rows, num_cols, 2), dtype=int)
            regions = np.empty(
                (num_rows, num_cols, assay.dims["channel"], self.region_length, self.region_length),
                dtype=assay.region.dtype,
            )
            for i in range(num_rows):
                for j in range(num_cols):
                    top, bottom, left, right = utils.bounding_box(
                        round(float(assay.x[i, j, t])),
                        round(float(assay.y[i, j, t])),
                        self.region_length,
                        assay.dims["im_row"],
                        assay.dims["im_col"],
                    )
                    regions[i, j] = images[:, top:bottom, left:right]
                    offsets[i, j] = [left, top]
            assay.region[:, :, :, t] = regions

            # Step 7: Compute the foreground and background masks for all buttons.
            fg = np.empty(
                (num_rows, num_cols, assay.dims["channel"], self.region_length, self.region_length),
                dtype=bool,
            )
            bg = np.empty_like(fg)
            for i in range(num_rows):
                for j in range(num_cols):
                    # TODO: This step should occur over multiple channels.
                    c = np.where(assay.channel == search_channels[0])[0][0]
                    subimage = utils.to_uint8(regions[i, j, c])
                    # Filter the subimage to smooth edges and remove noise.
                    filtered = cv.bilateralFilter(
                        subimage,
                        d=9,
                        sigmaColor=75,
                        sigmaSpace=75,
                        borderType=cv.BORDER_DEFAULT,
                    )

                    # Find any circles in the subimage.
                    circles = cv.HoughCircles(
                        filtered,
                        method=cv.HOUGH_GRADIENT,
                        dp=1,
                        minDist=50,
                        param1=20,
                        param2=5,
                        minRadius=self.min_button_radius,
                        maxRadius=self.max_button_radius,
                    )

                    # Update our estimate of the button position if we found some circles.
                    if circles is not None:
                        circles = circles[0, :, :2]
                        point = np.array([assay.x[i, j, t], assay.y[i, j, t]]) - offsets[i, j]
                        # Use the circle center closest to our previous estimate of the button.
                        closest_idx = np.argmin(np.linalg.norm(circles - point, axis=1))
                        assay.x[i, j, t], assay.y[i, j, t] = circles[closest_idx] + offsets[i, j]

                    x_rel = round(float(assay.x[i, j, t])) - offsets[i, j, 0]
                    y_rel = round(float(assay.y[i, j, t])) - offsets[i, j, 1]

                    # Set the foreground (the button) to be a circle of fixed radius.
                    fg_mask = utils.circle(
                        self.region_length,
                        row=y_rel,
                        col=x_rel,
                        radius=self.max_button_radius,
                        value=True,
                    )

                    # Set the background to be the annulus around our foreground.
                    bg_mask = utils.circle(
                        self.region_length,
                        row=y_rel,
                        col=x_rel,
                        radius=2 * self.max_button_radius,
                        value=True,
                    )
                    bg_mask &= ~fg_mask

                    # Refine the foreground & background by finding areas that are bright and dim.
                    _, bright_mask = cv.threshold(
                        subimage, thresh=0, maxval=1, type=cv.THRESH_BINARY + cv.THRESH_OTSU
                    )
                    dim_mask = ~cv.dilate(
                        bright_mask, np.ones((self.max_button_radius, self.max_button_radius))
                    )
                    bright_mask = bright_mask.astype(bool)
                    dim_mask = dim_mask.astype(bool)

                    # If part of the button is bright then set the foreground to that bright area.
                    if np.any(fg_mask & bright_mask):
                        fg_mask &= bright_mask

                    # The background on the other hand should not be bright.
                    if np.any(bg_mask & dim_mask):
                        bg_mask &= dim_mask

                    fg[i, j] = fg_mask
                    bg[i, j] = bg_mask

            assay.fg[:, :, :, t] = fg
            assay.bg[:, :, :, t] = bg

        return assay

    @registry.components.register("find_buttons")
    def make():
        return ButtonFinder()


class BeadFinder:
    def __init__(
        self,
        min_bead_radius: int = 10,
        max_bead_radius: int = 30,
        region_length: int = 61,
        param1: int = 50,
        param2: int = 30,
    ):
        self.min_bead_radius = min_bead_radius
        self.max_bead_radius = max_bead_radius
        self.region_length = region_length
        self.param1 = param1
        self.param2 = param2

    def __call__(
        self,
        assay: xr.Dataset,
        min_bead_radius: int = 10,
        max_bead_radius: int = 30,
        region_length: int = 61,
        param1: int = 50,
        param2: int = 30,
    ) -> xr.Dataset:
        self.min_bead_radius = min_bead_radius
        self.max_bead_radius = max_bead_radius
        self.region_length = region_length
        self.param1 = param1
        self.param2 = param2
        if isinstance(assay.search_channel, str):
            if assay.search_channel in assay.channel:
                search_channels = [assay.search_channel]
            elif assay.search_channel == "all":
                search_channels = assay.channel
            else:
                raise ValueError(f"{assay.search_channel} is not a channel name.")
        else:
            # We're searching across multiple channels.
            search_channels = assay.search_channel

        centers = np.empty((0, 2))
        radii = np.empty((0))
        for t in assay.time:
            for search_channel in search_channels:
                image = utils.to_uint8(assay.image.sel(channel=search_channel, time=t).to_numpy())
                # Filter the subimage to smooth edges and remove noise.
                filtered = cv.bilateralFilter(
                    image,
                    d=9,
                    sigmaColor=75,
                    sigmaSpace=75,
                    borderType=cv.BORDER_DEFAULT,
                )

                # Find any circles in the image.
                circles = cv.HoughCircles(
                    filtered,
                    method=cv.HOUGH_GRADIENT,
                    dp=1,
                    minDist=2 * self.min_bead_radius,
                    param1=self.param1,
                    param2=self.param2,
                    minRadius=self.min_bead_radius,
                    maxRadius=self.max_bead_radius,
                )

                if circles is not None:
                    circles = circles[0]
                    # Save circle center locations.
                    c = circles[:, :2]
                    if len(centers) > 0:
                        # Remove centers too close to those we already found in another channel or time.
                        dist_matrix = np.linalg.norm(c[np.newaxis] - centers[:, np.newaxis], axis=2)
                        valid = np.min(dist_matrix, axis=0) > self.min_bead_radius
                    else:
                        valid = np.ones(len(circles), dtype=bool)

                    centers = np.concatenate([centers, c[valid]])
                    radii = np.concatenate([radii, circles[valid, 2]])

        # Update the assay object with the beads we found.
        if len(centers) > 0:
            num_beads = len(centers)
            # Create the array of subimage regions.
            assay = assay.assign(
                region=(
                    ("marker", "channel", "time", "row", "col"),
                    np.empty(
                        (
                            num_beads,
                            assay.dims["channel"],
                            assay.dims["time"],
                            self.region_length,
                            self.region_length,
                        ),
                        dtype=assay.image.dtype,
                    ),
                ),
                fg=(
                    ("marker", "channel", "time", "row", "col"),
                    np.empty(
                        (
                            num_beads,
                            assay.dims["channel"],
                            assay.dims["time"],
                            self.region_length,
                            self.region_length,
                        ),
                        dtype=bool,
                    ),
                ),
                bg=(
                    ("marker", "channel", "time", "row", "col"),
                    np.empty(
                        (
                            num_beads,
                            assay.dims["channel"],
                            assay.dims["time"],
                            self.region_length,
                            self.region_length,
                        ),
                        dtype=bool,
                    ),
                ),
            )
            assay = assay.assign_coords(
                x=(
                    ["marker", "time"],
                    np.repeat(centers[:, np.newaxis, 0], assay.dims["time"], axis=1),
                ),
                y=(
                    ["marker", "time"],
                    np.repeat(centers[:, np.newaxis, 1], assay.dims["time"], axis=1),
                ),
            )

            # Compute the foreground and background masks for all buttons.
            for i in range(num_beads):
                # Set the subimage region for this bead.
                top, bottom, left, right = utils.bounding_box(
                    round(float(assay.x[i, 0])),
                    round(float(assay.y[i, 0])),
                    self.region_length,
                    image.shape[-2],
                    image.shape[-1],
                )
                assay.region[i] = assay.image.sel(
                    im_row=slice(top, bottom), im_col=slice(left, right)
                )

                # Set the foreground of the bead to be the circle we found.
                fg_mask = utils.circle(
                    self.region_length,
                    row=round(float(assay.y[i, 0] - top)),
                    col=round(float(assay.x[i, 0] - left)),
                    radius=round(radii[i]),
                    value=True,
                )

                # Set the background to be everything else in the region,
                # which could include other beads.
                bg_mask = ~fg_mask

                # Set the masked arrays with our computed masks.
                assay.fg[i] = fg_mask
                assay.bg[i] = bg_mask

        return assay

    @registry.components.register("find_beads")
    def make():
        return BeadFinder()


def cluster_1d(
    points: np.ndarray,
    total_length: int,
    num_clusters: int,
    cluster_length: float,
    ideal_num_points: np.ndarray,
    penalty: float,
) -> np.ndarray:
    # Find the best clustering using the accumulate ragged array trick.
    # See: https://vladfeinberg.com/2021/01/07/vectorizing-ragged-arrays.html
    permutation = np.argsort(points)
    points = points[permutation]

    def cost(offset: int) -> tuple[float, np.ndarray]:
        # Compute the boundaries and center of each cluster.
        boundaries = np.arange(num_clusters + 1) * cluster_length + offset
        centers = (boundaries[1:] + boundaries[:-1]) / 2
        # Find start/end indexes and number of points in each cluster.
        spans = np.searchsorted(points, boundaries)
        num_points = spans[1:] - spans[:-1]
        # Compute the squared distance of each point that falls in a cluster to its cluster center.
        dists = (points[spans[0] : spans[-1]] - np.repeat(centers, num_points)) ** 2
        # Penalize clusters for high variance.
        cost = np.insert(np.cumsum(dists), 0, 0)
        cost = np.diff(cost[spans - spans[0]])
        cost[num_points > 0] /= num_points[num_points > 0]
        # Set empty clusters to the maximum variance.
        cost[num_points == 0] = np.max(cost)
        # Fewer points in a cluster means noisier variances so adjust variance contributions.
        cost *= np.sqrt(ideal_num_points)
        # Penalize clusters for having too few or too many points.
        cost = cost + penalty * (ideal_num_points - num_points) ** 2
        return np.sum(cost), spans

    spans = min(
        (cost(i) for i in range(total_length - round(num_clusters * cluster_length))),
        key=lambda x: x[0],
    )[1]

    # Label each point with its cluster, label points outside clusters as -1.
    labels = -np.ones_like(points, dtype=int)
    labels[spans[0] : spans[-1]] = np.repeat(np.arange(num_clusters), spans[1:] - spans[:-1])

    # Return the labels based on the original order of the points.
    return labels[np.argsort(permutation)]


def regress_clusters(
    x: np.ndarray,
    y: np.ndarray,
    labels: np.ndarray,
    num_clusters: int,
    ideal_num_points: np.ndarray,
) -> tuple[np.ndarray, np.ndarray]:
    # Find the best line per-cluster.
    slopes = np.full(num_clusters, np.nan)
    intercepts = np.full(num_clusters, np.nan)
    cluster_points = [(x[labels == i], y[labels == i]) for i in range(num_clusters)]
    for i, (x, y) in enumerate(cluster_points):
        # Only regress on multi-point clusters.
        if len(x) > 1:
            slopes[i], intercepts[i], _, _, _ = scipy.stats.linregress(x, y)
        elif (i == 0 or i == num_clusters - 1) and ideal_num_points[i] >= 2:
            logger.warning(
                "Boundary cluster has fewer than 2 points."
                "The chip is unlikely to be segmented correctly."
            )

    # Recompute the intercepts using the median slope.
    slope = np.nanmedian(slopes)
    for i, (x, y) in enumerate(cluster_points):
        if len(x) > 0:
            intercepts[i] = np.median(y - slope * x)

    # Globally estimate where intercepts are, assuming they're evenly spaced.
    not_nan = ~np.isnan(intercepts)
    label_idxs = np.arange(num_clusters)
    intercept_m, intercept_b, _, _, _ = scipy.stats.linregress(
        label_idxs[not_nan], intercepts[not_nan]
    )
    # Re-estimate intercepts using a weighted mean of global and local estimates.
    # This reduces outlier effects while still allowing uneven intercepts from image stitching.
    for i, (x, y) in enumerate(cluster_points):
        if ideal_num_points[i] != 0 and not_nan[i]:
            weight = min(len(x), ideal_num_points[i]) / ideal_num_points[i]
            intercepts[i] = weight * intercepts[i] + (1 - weight) * (intercept_m * i + intercept_b)
        else:
            # Just use our global estimate when we have an cluster.
            intercepts[i] = intercept_m * i + intercept_b

    return slope, intercepts



from __future__ import annotations
from typing import Callable

from numpy.typing import ArrayLike
import catalogue

from magnify.pipeline import Pipeline

readers = catalogue.create("magnify", "readers")
components = catalogue.create("magnify", "components")


def load(name: str) -> Pipeline:
    if name == "chip":
        return chip_pipeline()
    elif name == "mrbles":
        return mrbles_pipeline()
    elif name == "beads":
        return mrbles_pipeline()
    elif name == "imread":
        return imread_pipeline()
    else:
        raise ValueError(f"Pipeline {name} does not exist.")


def chip_pipeline():
    pipe = Pipeline("read")
    pipe.add_pipe("read_pinlist")
    # pipe.add_pipe("preprocessor")
    pipe.add_pipe("flip_horizontal")
    pipe.add_pipe("stitch")
    pipe.add_pipe("find_buttons")
    # pipe.add_pipe("background_filter")

    return pipe


def mrbles_pipeline():
    pipe = Pipeline("read")
    pipe.add_pipe("flatfield_correct")
    pipe.add_pipe("stitch")
    pipe.add_pipe("find_beads")
    # pipe.add_pipe("background_filter")

    return pipe


def imread_pipeline():
    pipe = Pipeline("read")
    # pipe.add_pipe("flip_horizontal")
    # pipe.add_pipe("stitch")
    return pipe



from __future__ import annotations
from collections.abc import Callable
from typing import Any
import inspect
import re

import cv2 as cv
import numpy as np


def to_uint8(arr: np.ndarray) -> np.ndarray:
    arr = arr.astype(float)
    arr = 255 * arr / np.max(arr)
    return arr.astype(np.uint8)


def circle(image_length: int, row: int, col: int, radius: int, value: Any = 1.0) -> np.ndarray:
    image = np.zeros((image_length, image_length), dtype=np.uint8)
    cv.circle(image, (col, row), radius, 1, -1)
    image = image.astype(type(value)) * value
    return image


def ceildiv(a: int, b: int) -> int:
    return -(a // -b)


def bounding_box(
    x: int, y: int, box_length: int, image_width: int, image_height: int
) -> tuple[int, int, int, int]:
    top = y - box_length // 2
    bottom = y + ceildiv(box_length, 2)
    if top < 0:
        bottom += -top
        top = 0
    if bottom > image_height:
        top -= bottom - image_height
        bottom = image_height
    left = x - box_length // 2
    right = x + ceildiv(box_length, 2)
    if left < 0:
        right += -left
        left = 0
    if right > image_width:
        left -= right - image_width
        right = image_width
    return top, bottom, left, right


def valid_kwargs(kwargs: dict[str, Any], func: Callable) -> dict[str, Any]:
    args = list(inspect.signature(func).parameters)
    return {k: kwargs[k] for k in kwargs if k in args}


def natural_sort_key(s: str) -> list[str]:
    reg = re.compile("([0-9]+)")
    return [int(text) if text.isdigit() else text.lower() for text in reg.split(s)]



from __future__ import annotations
from typing import cast
import os

import basicpy
import tifffile
import numpy as np
import xarray as xr

import magnify.registry as registry


class Preprocessor:
    def __call__(self, assay: xr.Dataset) -> xr.Dataset:
        tiles = assay.image
        flat_tiles = assay.image.reshape(-1, tiles.dim["im_row"], tiles.dim["im_col"])
        model = basicpy.basicpy.BaSiC(get_darkfield=True, smoothness_flatfield=1)
        result = cast(np.ndarray, model.fit_transform(flat_tiles, timelapse=False))
        assay = assay.assign(image=result.reshape(*tiles.shape))
        return assay

    @registry.components.register("preprocess")
    def make():
        return Preprocessor()


@registry.components.register("flatfield_correct")
def make_flatfield_correct():
    def flatfield_correct(assay: xr.Dataset, flatfield=1.0, darkfield=0.0):
        if isinstance(flatfield, str):
            with tifffile.TiffFile(os.path.expanduser(flatfield)) as tif:
                flatfield = tif.asarray()
        if isinstance(darkfield, str):
            with tifffile.TiffFile(os.path.expanduser(darkfield)) as tif:
                darkfield = tif.asarray()

        assay["image"] = (assay.image - darkfield) / flatfield
        return assay

    return flatfield_correct


@registry.components.register("flip_horizontal")
def make_horizontal_flip():
    def horizontal_flip(assay: xr.Dataset):
        assay["image"] = assay.image.isel(im_col=slice(None, None, -1))
        return assay

    return horizontal_flip



from __future__ import annotations
from collections.abc import Iterator
import collections
import datetime
import fnmatch
import functools
import glob
import os
import re

from numpy.typing import ArrayLike
from typing import Iterator
import bs4
import dask.array as da
import numpy as np
import pandas as pd
import tifffile
import xarray as xr

from magnify.pipeline import Pipeline
import magnify.registry as registry
import magnify.utils as utils


class Reader:
    def __init__(self) -> None:
        pass

    def __call__(
        self,
        data: ArrayLike | str,
        names: ArrayLike | str,
        search_on: str = "all",
        times: Sequence[int] | None = None,
        channels: Sequence[str] | None = None,
    ) -> Iterator[xr.Dataset]:
        if isinstance(data, str):
            data = [data]

        for d in data:
            path_dict = extract_paths(d)
            if len(path_dict) == 0:
                raise FileNotFoundError(f"The pattern {d} did not lead to any files.")

            for assay_name, assay_dict in sorted(
                path_dict.items(), key=lambda x: utils.natural_sort_key(x[0])
            ):
                # Use these variables within the loop so we don't affect other assays.
                channel_coords = channels
                time_coords = times

                # Get the indices specified in the path each in sorted order.
                channel_idxs, time_idxs, row_idxs, col_idxs = (
                    sorted(set(idx)) for idx in zip(*assay_dict.keys())
                )

                # Check which dimensions are specified in the path and compute the shape of those dimensions.
                dims_in_path = []
                outer_shape = tuple()
                if channel_idxs[0] != -1:
                    dims_in_path.append("channel")
                    outer_shape += (len(channel_idxs),)
                if time_idxs[0] != -1:
                    dims_in_path.append("time")
                    outer_shape += (len(time_idxs),)
                if row_idxs[0] != -1:
                    dims_in_path.append("tile_row")
                    outer_shape += (len(row_idxs),)
                if col_idxs[0] != -1:
                    dims_in_path.append("tile_col")
                    outer_shape += (len(col_idxs),)

                # If the user didn't specify times or channels use the ones from the path.
                if time_coords is None and "time" in dims_in_path:
                    time_coords = time_idxs
                if channel_coords is None and "channel" in dims_in_path:
                    channel_coords = channel_idxs

                # Read in a single image to get the metadata stored within the file.
                with tifffile.TiffFile(next(iter(assay_dict.values()))) as tif:
                    dtype = tif.series[0].dtype
                    inner_shape = tif.series[0].shape
                    page_shape = tif.pages[0].shape

                    letter_to_dim = {
                        "C": "channel",
                        "T": "time",
                        "Z": "depth",
                        "Y": "im_row",
                        "X": "im_col",
                        "R": "tile_pos",
                    }
                    dims_in_file = [letter_to_dim[c] for c in tif.series[0].axes]

                    if (
                        time_coords is None
                        and tif.is_micromanager
                        and "StartTime" in tif.micromanager_metadata["Summary"]
                    ):
                        # Get the time string without timezone info.
                        time_str = tif.micromanager_metadata["Summary"]["StartTime"][:-6]
                        start_time = datetime.datetime.strptime(time_str, "%Y-%m-%d %H:%M:%S.%f")
                        if "time" in dims_in_file:
                            # Only look at the first file's description since time isn't across multiple files.
                            planes = [
                                pl
                                for pl in bs4.BeautifulSoup(tif.pages[0].description, "xml")
                                .find("Image")
                                .find_all("Plane")
                            ]
                            assert all(pl.get("DeltaTUnit") == "ms" for pl in planes)
                            time_coords = [
                                start_time
                                + datetime.timedelta(milliseconds=float(pl.get("DeltaT")))
                                for pl in planes
                            ]
                            if "channel" in dims_in_file:
                                stride = inner_shape[dims_in_file.index("channel")]
                            else:
                                stride = 1
                            assert len(time_coords) % stride == 0
                            time_coords = time_coords[::stride]
                        else:
                            time_coords = [start_time]
                    if channel_coords is None and tif.is_micromanager:
                        channel_coords = tif.micromanager_metadata["Summary"]["ChNames"]

                    if "tile_pos" in dims_in_file:
                        # Tiles are always saved in multiple files so ignore this dimension
                        # since the user should specify tiles in their search path.
                        tile_idx = dims_in_file.index("tile_pos")
                        inner_shape = inner_shape[:tile_idx] + inner_shape[tile_idx + 1 :]
                        dims_in_file = dims_in_file[:tile_idx] + dims_in_file[tile_idx + 1 :]

                    if "depth" in dims_in_file:
                        raise ValueError("tiff files with a Z dimension are not yet supported.")
                    if "im_row" not in dims_in_file or "im_col" not in dims_in_file:
                        raise ValueError("tiff files must contain an X and Y dimension.")

                # Check the dimensions specified in the path and inside the tiff file do not overlap.
                if set(dims_in_file).intersection(dims_in_path):
                    raise ValueError(
                        "Dimensions specified in the path names and inside the tiff file overlap."
                    )

                # Setup a dask array to lazyload images.
                filenames = [path for _, path in sorted(assay_dict.items())]

                def read_image(block_id, filenames):
                    outer_id = block_id[: len(outer_shape)]
                    inner_id = block_id[len(outer_shape) :]
                    if len(outer_id) > 0:
                        file_idx = np.ravel_multi_index(outer_id, outer_shape)
                    else:
                        # This is the case where we don't have indices outside the file.
                        file_idx = 0

                    with tifffile.TiffFile(filenames[file_idx]) as tif:
                        page_ndim = len(page_shape)
                        if len(inner_shape) > page_ndim:
                            page_idx = np.ravel_multi_index(
                                inner_id[:-page_ndim], inner_shape[:-page_ndim]
                            )
                        else:
                            # This is the case where no dimensions correspond to page dims.
                            page_idx = 0
                        # Read the page from disk.
                        page = tif.pages[page_idx].asarray()
                        # Expand the dimensions of the block to incorporate outer dims and page index dims.
                        return np.expand_dims(page, axis=tuple(range(len(block_id) - page_ndim)))

                # Chunk the images so that each chunk represents a single page in the tiff file.
                images = da.map_blocks(
                    functools.partial(read_image, filenames=filenames),
                    dtype=dtype,
                    chunks=(
                        (
                            tuple((1,) * size for size in outer_shape)
                            + tuple((1,) * size for size in inner_shape[: -len(page_shape)])
                            + inner_shape[-len(page_shape) :]
                        )
                    ),
                )

                # Set named coordinates for channels and times if they're available.
                coords = {}
                if channel_coords is not None:
                    coords["channel"] = channel_coords
                if time_coords is not None:
                    coords["time"] = time_coords

                # Put all our data into an xarray dataset.
                assay = xr.Dataset(
                    {"image": (dims_in_path + dims_in_file, images)},
                    coords=coords,
                    attrs={
                        "name": assay_name,
                        "search_channel": search_on,
                    },
                )

                # Make sure the assay always has a time and channel dimension.
                if "channel" not in assay.image.dims:
                    assay["image"] = assay.image.expand_dims("channel", 0)
                if "time" not in assay.image.dims:
                    assay["image"] = assay.image.expand_dims("time", 1)

                # Reorder the dimensions so they're always consistent.
                desired_order = []
                for dim in ["channel", "time", "tile_row", "tile_col", "im_row", "im_col"]:
                    if dim in assay.dims:
                        desired_order.append(dim)
                assay = assay.transpose(*desired_order)

                yield assay

    @registry.readers.register("read")
    def make():
        return Reader()


@registry.components.register("read_pinlist")
def make_read_pinlist():
    def read_pinlist(assay, pinlist):
        assay = assay.assign_coords(id=(("marker_row", "marker_col"), read_names(pinlist)))
        return assay

    return read_pinlist


def read_names(path):
    df = pd.read_csv(path)
    df["Indices"] = df["Indices"].apply(
        lambda s: [int(x) for x in re.sub(r"[\(\)]", "", s).split(",")]
    )
    # Zero-index the indices.
    cols, rows = np.array(df["Indices"].to_list()).T - 1
    names = df["MutantID"].to_numpy(dtype=str, na_value="")
    names_array = np.empty((max(rows) + 1, max(cols) + 1), dtype=names.dtype)
    names_array[rows, cols] = names
    return names_array


def extract_paths(pattern) -> dict[tuple[int, str, int, int], str]:
    pattern = os.path.expanduser(pattern)

    # Filter out special signifiers used for pattern matching.
    glob_path = pattern.replace("(assay)", "*")
    glob_path = glob_path.replace("(channel)", "*")
    glob_path = re.sub(r"\(time\s*\|?.*?\)", "*", glob_path)
    glob_path = glob_path.replace("(row)", "*")
    glob_path = glob_path.replace("(col)", "*")

    # Search for files matching the pattern.
    paths = glob.glob(glob_path, recursive=True)

    regex_path = fnmatch.translate(pattern)
    regex_path = regex_path.replace("\\(assay\\)", "(?P<assay>.*)")
    regex_path = regex_path.replace("\\(channel\\)", "(?P<channel>.*)")
    regex_path = re.sub(r"\\\(time\s*\|?.*?\\\)", r"(?P<time>.*)", regex_path)
    regex_path = regex_path.replace("\\(row\\)", "(?P<row>.*)")
    regex_path = regex_path.replace("\\(col\\)", "(?P<col>.*)")
    regex_path = re.compile(regex_path, re.IGNORECASE)

    path_dict = collections.defaultdict(dict)
    for path in paths:
        match = regex_path.fullmatch(path)
        if "(assay)" in pattern:
            assay = match.group("assay")
        else:
            assay = ""

        if "(channel)" in pattern:
            channel = match.group("channel")
        else:
            channel = -1

        time_search = re.search(r"\(time\s*\|?\s*(.*?)\)", pattern)
        if time_search:
            format_str = time_search.group(1)
            if not format_str:
                format_str = "%Y%m%d-%H%M%S"
            time_str = match.group("time")
            time = datetime.datetime.strptime(time_str, format_str)
        else:
            time = -1

        if "(row)" in pattern:
            row = int(match.group("row"))
        else:
            row = -1

        if "(col)" in pattern:
            col = int(match.group("col"))
        else:
            col = -1

        idx = (channel, time, row, col)
        if idx not in path_dict[assay]:
            path_dict[assay][idx] = os.path.abspath(path)
        else:
            raise ValueError(f"{path} and {path_dict[assay][idx]} map to the same index.")

    return path_dict



from __future__ import annotations
from typing import Callable

from numpy.typing import ArrayLike
import xarray as xr

import magnify.registry as registry
import magnify.utils as utils


class Pipeline:
    def __init__(self, reader: str):
        self.reader: Callable[[ArrayLike | str], xr.Dataset] = registry.readers.get(reader)()
        self.components: list[Callable[[xr.Dataset], xr.Dataset]] = []

    def __call__(
        self,
        data: ArrayLike | str,
        names: ArrayLike | str = None,
        search_on: str = "egfp",
        times: Sequence[int] | None = None,
        channels: Sequence[str] | None = None,
        **kwargs,
    ) -> xr.Dataset | list[xr.Dataset]:
        inputs = self.reader(
            data=data, names=names, search_on=search_on, times=times, channels=channels
        )
        assays = []
        for assay in inputs:
            for name, component in self.components:
                assay = component(assay, **utils.valid_kwargs(kwargs, component))

            assays.append(assay)

        if len(assays) == 1:
            assays = assays[0]

        return assays

    def add_pipe(self, name: str) -> None:
        self.components.append((name, registry.components.get(name)()))
